"""This module is the frontend to the SigmaDSP backend service.

It can control the backend via the command line.
"""
import argparse
import logging
from typing import Union

import grpc

from sigmadsp.generated.backend_service.control_pb2 import (
    ControlParameterRequest,
    ControlRequest,
    ControlResponse,
)
from sigmadsp.generated.backend_service.control_pb2_grpc import BackendStub


def main():
    """The main frontend command-line application, which controls the SigmaDSP backend."""
    logging.basicConfig(level=logging.INFO)

    argument_parser = argparse.ArgumentParser()

    argument_parser.add_argument(
        "-p",
        "--port",
        required=False,
        type=int,
        help="Set the port, on which the backend listens for requests.",
    )

    argument_parser.add_argument(
        "-a",
        "--address",
        required=False,
        type=str,
        help="Set the IP address, on which the backend listens for requests.",
    )

    argument_parser.add_argument(
        "-av",
        "--adjust_volume",
        required=False,
        type=float,
        help="Adjust the volume by a certain value in dB (positive or negative).",
    )

    argument_parser.add_argument(
        "-sv",
        "--set_volume",
        required=False,
        type=float,
        help="Sets the volume to a certain value in dB (zero or lower).",
    )

    argument_parser.add_argument(
        "-r",
        "--reset",
        required=False,
        help="Soft-reset the DSP.",
        action="store_true",
    )

    argument_parser.add_argument(
        "-R",
        "--hard-reset",
        required=False,
        help="hard-reset the DSP.",
        action="store_true",
    )

    argument_parser.add_argument(
        "-lp",
        "--load_parameters",
        required=False,
        help="Load new parameter file",
    )

    arguments = argument_parser.parse_args()

    backend_port = 50051
    backend_address = "localhost"

    if arguments.port is not None:
        backend_port = arguments.port

    if arguments.address is not None:
        backend_address = arguments.address

    response: Union[ControlResponse, None] = None
    control_request = ControlRequest()
    control_parameter_request = ControlParameterRequest()

    with grpc.insecure_channel(f"{backend_address}:{backend_port}") as channel:
        stub = BackendStub(channel)

        if arguments.adjust_volume is not None:
            control_parameter_request.change_volume.name_tokens[:] = ["main"]
            control_parameter_request.change_volume.value = arguments.adjust_volume
            control_parameter_request.change_volume.relative = True

            response = stub.control_parameter(control_parameter_request)

        if arguments.set_volume is not None:
            control_parameter_request.change_volume.name_tokens[:] = ["main"]
            control_parameter_request.change_volume.value = arguments.set_volume
            control_parameter_request.change_volume.relative = False

            response = stub.control_parameter(control_parameter_request)

        if arguments.load_parameters is not None:
            with open(arguments.load_parameters, "r", encoding="utf8") as parameter_file:
                control_request.load_parameters.content[:] = parameter_file.readlines()

            response = stub.control(control_request)

        if arguments.reset is True:
            control_request.reset_dsp = True

            response = stub.control(control_request)

        if arguments.hard_reset is True:
            control_request.hard_reset_dsp = True

            response = stub.control(control_request)

        logging.info(response and response.message)


if __name__ == "__main__":
    main()



"""This module includes many conversion functions that can be used for different purposes.

- Conversion between linear and dB-scale values
- Conversion from bytes-like objects of varying length to integers
- Conversion from integers to bytes-like objects of specified length
"""
import math


def clamp(value: float, min_value: float, max_value: float) -> float:
    """Clamps a value to a specified range.

    Args:
        value (float): The value to clamp.
        min_value (float): Lower clamping boundary.
        max_value (float): Upper clamping boundary.

    Returns:
        float: The clamped value.
    """
    if value >= max_value:
        value = max_value

    elif value <= min_value:
        value = min_value

    return value


def frac_8_24_to_float(value: int) -> float:
    """Convert a value in the DSPs fractional representation to float.

    32 bit values consist of 8 integer and 24 fractional bits and are signed.

    Args:
        value (int): Fractional value to convert

    Returns:
        float: Output in float format
    """
    return value / 2**24


def float_to_frac_8_24(value: float) -> int:
    """Convert a float value to the DSPs fractional representation.

    32 bit values consist of 8 integer and 24 fractional bits and are signed.

    Args:
        value (float): Float value to convert

    Returns:
        int: Output in DSP fractional format
    """
    return int(value * 2**24)


def db_to_linear(value_db: float) -> float:
    """Convert a dB-scale value (e.g. voltage) to a linear-scale value.

    Args:
        value (float): Input dB value

    Returns:
        float: Output linear value
    """
    return 10 ** (value_db / 20)


def linear_to_db(value_linear: float) -> float:
    """Convert a linear-scale value to a dB-scale value (e.g. voltage).

    Args:
        value_linear (float): The linear input value

    Returns:
        float: Output in dB scale
    """
    return 20 * math.log10(value_linear)


def bytes_to_int(data: bytes, offset: int = 0, length: int = 1) -> int:
    """Convert a number of bytes to their integer representation.

    Uses "length" bytes from the "data" input, starting at "offset".

    Args:
        data (bytes): Input bytes
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
        length (int, optional): Number of bytes to convert. Defaults to 1.

    Returns:
        int: Integer representation of the input data stream
    """
    return int.from_bytes(data[offset : offset + length], byteorder="big")


def bytes_to_int8(data: bytes, offset: int = 0) -> int:
    """Convert one byte to an 8 bit integer value.

    Args:
        data (bytes): Input byte
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer

    Returns:
        int: 8 bit integer representation of the input data stream
    """
    return bytes_to_int(data, offset, length=1)


def bytes_to_int16(data: bytes, offset: int = 0) -> int:
    """Convert two bytes to a 16 bit integer value.

    Args:
        data (bytes): Input bytes
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer

    Returns:
        int: 16 bit integer representation of the input data stream
    """
    return bytes_to_int(data, offset, length=2)


def bytes_to_int32(data: bytes, offset: int = 0) -> int:
    """Convert four bytes to a 32 bit integer value.

    Args:
        data (bytes): Input bytes
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer

    Returns:
        int: 32 bit integer representation of the input data stream
    """
    return bytes_to_int(data, offset, length=4)


def int_to_bytes(value: int, buffer: bytearray = None, offset: int = 0, length: int = 1):
    """Fill a buffer with values. If no buffer is provided, a new one is created.

    Args:
        buffer (bytearray): The buffer to fill
        value (int): The value to pack into the buffer
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
        length (int): Number of bytes to be written
    """
    if buffer is None:
        buffer = bytearray(length + offset)

    buffer[offset : offset + length] = value.to_bytes(length, byteorder="big")

    return buffer


def int8_to_bytes(value, buffer=None, offset=0):
    """Fill a buffer with an 8 bit value (1 byte). If no buffer is provided, a new one is created.

    Args:
        buffer (bytearray): The buffer to fill
        value (int): The value to pack into the buffer
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
    """
    return int_to_bytes(value, buffer, offset=offset, length=1)


def int16_to_bytes(value, buffer=None, offset=0):
    """Fill a buffer with a 16 bit value (2 bytes). If no buffer is provided, a new one is created.

    Args:
        value (int): The value to pack into the buffer
        buffer (bytearray, optional): The buffer to fill
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
    """
    return int_to_bytes(value, buffer, offset=offset, length=2)


def int32_to_bytes(value, buffer=None, offset=0) -> bytearray:
    """Fill a buffer with a 32 bit value (4 bytes). If no buffer is provided, a new one is created.

    Args:
        buffer (bytearray): The buffer to fill
        value (int): The value to pack into the buffer
        offset (int, optional): Offset in number of bytes, from the beginning of the data buffer
    """
    return int_to_bytes(value, buffer, offset=offset, length=4)



"""This module implements the base class for I2C and SPI communication handlers."""
import logging
import threading
from abc import ABC, abstractmethod
from multiprocessing import Pipe

# A logger for this module
logger = logging.getLogger(__name__)


class BaseProtocol(ABC):
    """Base class for communication handlers talking to SigmaDSP chipsets."""

    def __init__(self, bus: int = 0, device: int = 0):
        """Initialize the communications thread.

        Args:
            bus (int, optional): Bus number. Defaults to 0.
            device (int, optional): Device number / address. Defaults to 0
        """
        self._initialize(bus, device)

        # Generate a Pipe, for communicating with the protocol handler thread within this class.
        self.pipe_end_owner, self.pipe_end_user = Pipe()

        protocol = self.__class__.__name__

        logger.info("Starting %s handling thread.", protocol)
        self.thread = threading.Thread(target=self.serve_forever, name=f"{protocol} handler thread", daemon=True)
        self.thread.start()

    def write(self, address: int, data: bytes):
        """Write data over the hardware interface via the pipe to the hardware thread.

        Args:
            address (int): DSP register address to write to
            data (bytes): Binary data to write
        """
        self.pipe_end_owner.send("write")
        self.pipe_end_owner.send((address, data))

    def read(self, address: int, length: int) -> bytes:
        """Read data from the hardware interface via the pipe to the hardware thread.

        Args:
            address (int): DSP register address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Register content
        """
        self.pipe_end_owner.send("read")
        self.pipe_end_owner.send((address, length))

        data = self.pipe_end_owner.recv()

        return data

    def serve_forever(self):
        """Handle incoming requests for writing or reading data."""
        while True:
            mode = self.pipe_end_user.recv()

            if mode == "write":
                address, data = self.pipe_end_user.recv()
                self._write(address, data)

            elif mode == "read":
                address, length = self.pipe_end_user.recv()
                data = self._read(address, length)
                self.pipe_end_user.send(data)

    @abstractmethod
    def _initialize(self, bus: int = 0, device: int = 0):
        """Initialize the hardware.

        Args:
            bus (int, optional): Bus number. Defaults to 0.
            device (int, optional): Device number / address. Defaults to 0
        """

    @abstractmethod
    def _read(self, address: int, length: int) -> bytes:
        """Read data from a SigmaDSP.

        Args:
            address (int): Address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Data that was read from the DSP
        """

    @abstractmethod
    def _write(self, address: int, data: bytes):
        """Write data onto a SigmaDSP.

        Args:
            address (int): Address to write to
            data (bytes): Data to write
        """



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: control.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='control.proto',
  package='sigmadsp.backend_service',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\rcontrol.proto\x12\x18sigmadsp.backend_service\"D\n\x0c\x43hangeVolume\x12\x13\n\x0bname_tokens\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x10\n\x08relative\x18\x03 \x01(\x08\"e\n\x17\x43ontrolParameterRequest\x12?\n\rchange_volume\x18\x01 \x01(\x0b\x32&.sigmadsp.backend_service.ChangeVolumeH\x00\x42\t\n\x07\x63ommand\"!\n\x0eLoadParameters\x12\x0f\n\x07\x63ontent\x18\x01 \x03(\t\"\x8f\x01\n\x0e\x43ontrolRequest\x12\x13\n\treset_dsp\x18\x01 \x01(\x08H\x00\x12\x18\n\x0ehard_reset_dsp\x18\x02 \x01(\x08H\x00\x12\x43\n\x0fload_parameters\x18\x03 \x01(\x0b\x32(.sigmadsp.backend_service.LoadParametersH\x00\x42\t\n\x07\x63ommand\"3\n\x0f\x43ontrolResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t2\xdc\x01\n\x07\x42\x61\x63kend\x12^\n\x07\x63ontrol\x12(.sigmadsp.backend_service.ControlRequest\x1a).sigmadsp.backend_service.ControlResponse\x12q\n\x11\x63ontrol_parameter\x12\x31.sigmadsp.backend_service.ControlParameterRequest\x1a).sigmadsp.backend_service.ControlResponseb\x06proto3'
)




_CHANGEVOLUME = _descriptor.Descriptor(
  name='ChangeVolume',
  full_name='sigmadsp.backend_service.ChangeVolume',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name_tokens', full_name='sigmadsp.backend_service.ChangeVolume.name_tokens', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sigmadsp.backend_service.ChangeVolume.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='relative', full_name='sigmadsp.backend_service.ChangeVolume.relative', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=43,
  serialized_end=111,
)


_CONTROLPARAMETERREQUEST = _descriptor.Descriptor(
  name='ControlParameterRequest',
  full_name='sigmadsp.backend_service.ControlParameterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='change_volume', full_name='sigmadsp.backend_service.ControlParameterRequest.change_volume', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='command', full_name='sigmadsp.backend_service.ControlParameterRequest.command',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=113,
  serialized_end=214,
)


_LOADPARAMETERS = _descriptor.Descriptor(
  name='LoadParameters',
  full_name='sigmadsp.backend_service.LoadParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='content', full_name='sigmadsp.backend_service.LoadParameters.content', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=216,
  serialized_end=249,
)


_CONTROLREQUEST = _descriptor.Descriptor(
  name='ControlRequest',
  full_name='sigmadsp.backend_service.ControlRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='reset_dsp', full_name='sigmadsp.backend_service.ControlRequest.reset_dsp', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hard_reset_dsp', full_name='sigmadsp.backend_service.ControlRequest.hard_reset_dsp', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='load_parameters', full_name='sigmadsp.backend_service.ControlRequest.load_parameters', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='command', full_name='sigmadsp.backend_service.ControlRequest.command',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=252,
  serialized_end=395,
)


_CONTROLRESPONSE = _descriptor.Descriptor(
  name='ControlResponse',
  full_name='sigmadsp.backend_service.ControlResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='sigmadsp.backend_service.ControlResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='sigmadsp.backend_service.ControlResponse.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=397,
  serialized_end=448,
)

_CONTROLPARAMETERREQUEST.fields_by_name['change_volume'].message_type = _CHANGEVOLUME
_CONTROLPARAMETERREQUEST.oneofs_by_name['command'].fields.append(
  _CONTROLPARAMETERREQUEST.fields_by_name['change_volume'])
_CONTROLPARAMETERREQUEST.fields_by_name['change_volume'].containing_oneof = _CONTROLPARAMETERREQUEST.oneofs_by_name['command']
_CONTROLREQUEST.fields_by_name['load_parameters'].message_type = _LOADPARAMETERS
_CONTROLREQUEST.oneofs_by_name['command'].fields.append(
  _CONTROLREQUEST.fields_by_name['reset_dsp'])
_CONTROLREQUEST.fields_by_name['reset_dsp'].containing_oneof = _CONTROLREQUEST.oneofs_by_name['command']
_CONTROLREQUEST.oneofs_by_name['command'].fields.append(
  _CONTROLREQUEST.fields_by_name['hard_reset_dsp'])
_CONTROLREQUEST.fields_by_name['hard_reset_dsp'].containing_oneof = _CONTROLREQUEST.oneofs_by_name['command']
_CONTROLREQUEST.oneofs_by_name['command'].fields.append(
  _CONTROLREQUEST.fields_by_name['load_parameters'])
_CONTROLREQUEST.fields_by_name['load_parameters'].containing_oneof = _CONTROLREQUEST.oneofs_by_name['command']
DESCRIPTOR.message_types_by_name['ChangeVolume'] = _CHANGEVOLUME
DESCRIPTOR.message_types_by_name['ControlParameterRequest'] = _CONTROLPARAMETERREQUEST
DESCRIPTOR.message_types_by_name['LoadParameters'] = _LOADPARAMETERS
DESCRIPTOR.message_types_by_name['ControlRequest'] = _CONTROLREQUEST
DESCRIPTOR.message_types_by_name['ControlResponse'] = _CONTROLRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ChangeVolume = _reflection.GeneratedProtocolMessageType('ChangeVolume', (_message.Message,), {
  'DESCRIPTOR' : _CHANGEVOLUME,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ChangeVolume)
  })
_sym_db.RegisterMessage(ChangeVolume)

ControlParameterRequest = _reflection.GeneratedProtocolMessageType('ControlParameterRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLPARAMETERREQUEST,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ControlParameterRequest)
  })
_sym_db.RegisterMessage(ControlParameterRequest)

LoadParameters = _reflection.GeneratedProtocolMessageType('LoadParameters', (_message.Message,), {
  'DESCRIPTOR' : _LOADPARAMETERS,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.LoadParameters)
  })
_sym_db.RegisterMessage(LoadParameters)

ControlRequest = _reflection.GeneratedProtocolMessageType('ControlRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLREQUEST,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ControlRequest)
  })
_sym_db.RegisterMessage(ControlRequest)

ControlResponse = _reflection.GeneratedProtocolMessageType('ControlResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLRESPONSE,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:sigmadsp.backend_service.ControlResponse)
  })
_sym_db.RegisterMessage(ControlResponse)



_BACKEND = _descriptor.ServiceDescriptor(
  name='Backend',
  full_name='sigmadsp.backend_service.Backend',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=451,
  serialized_end=671,
  methods=[
  _descriptor.MethodDescriptor(
    name='control',
    full_name='sigmadsp.backend_service.Backend.control',
    index=0,
    containing_service=None,
    input_type=_CONTROLREQUEST,
    output_type=_CONTROLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='control_parameter',
    full_name='sigmadsp.backend_service.Backend.control_parameter',
    index=1,
    containing_service=None,
    input_type=_CONTROLPARAMETERREQUEST,
    output_type=_CONTROLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BACKEND)

DESCRIPTOR.services_by_name['Backend'] = _BACKEND

# @@protoc_insertion_point(module_scope)



"""This module implements an I2C handler that talks to Sigma DSP devices."""
import logging

from smbus2 import SMBus, i2c_msg

from sigmadsp.hardware.base_protocol import BaseProtocol
from sigmadsp.helper.conversion import int16_to_bytes

# A logger for this module
logger = logging.getLogger(__name__)


class I2C(BaseProtocol):
    """Handle I2C transfers from and to SigmaDSP chipsets.

    Tested with ADAU1701
    """

    def _initialize(self, bus: int = 1, device: int = 0x38):
        """Initialize the I2C hardware.

        Bus will be 1 for RaspberryPi hardware.

        Args:
            bus (int, optional): Bus number. Defaults to 1.
            device (int, optional): Device number. Defaults to 0x38. (ADAU145x default address)
        """
        self.bus = SMBus(bus)
        self.i2c_addr = device

    def _read(self, address: int, length: int) -> bytes:
        """Read data over the i2c port from a SigmaDSP.

        Args:
            address (int): Address to read from
            length (int): Number of bytes to read

        Returns:
            bytes: Data that was read from the DSP
        """
        address_bytes = list(int16_to_bytes(address))

        msg_wr = i2c_msg.write(self.i2c_addr, address_bytes)
        msg_rd = i2c_msg.read(self.i2c_addr, length)

        self.bus.i2c_rdwr(msg_wr, msg_rd)
        rd_list = bytes(msg_rd)

        return rd_list

    def _write(self, address: int, data: bytes):
        """Write data over the I2C port onto a SigmaDSP.

        Args:
            address (int): Address to write to
            data (bytes): Data to write
        """
        wr_content = list(int16_to_bytes(address))
        wr_content.extend(list(data))

        msg_wr = i2c_msg.write(self.i2c_addr, wr_content)

        self.bus.i2c_rdwr(msg_wr)



"""This module provides functionality for controlling SigmaDSP hardware, e.g.

- Deploying programs
- Changing parameter register contents
- Reading parameter registers
- Performing soft reset
"""
import logging
from typing import Union

from sigmadsp.hardware.dsp import Dsp
from sigmadsp.helper.conversion import (
    bytes_to_int32,
    clamp,
    db_to_linear,
    float_to_frac_8_24,
    frac_8_24_to_float,
    int16_to_bytes,
    int32_to_bytes,
    linear_to_db,
)

# A logger for this module
logger = logging.getLogger(__name__)


class Adau14xx(Dsp):
    """A class for controlling functionality of Analog Devices Sigma DSPs, especially ADAU14xx series parts."""

    # Addresses and sizes of important registers
    RESET_REGISTER = 0xF890
    RESET_REGISTER_LENGTH = 2

    # safeload registers - SigmaStudio 3.14 and newer
    SAFELOAD_DATA_REGISTERS = [0x6000, 0x6001, 0x6002, 0x6003, 0x6004]
    SAFELOAD_ADDRESS_REGISTER = 0x6005
    SAFELOAD_COUNT_REGISTER = 0x6006
    SAFELOAD_DATA_REGISTER_LENGTH = 4

    # All fixpoint (parameter) registers are four bytes long
    FIXPOINT_REGISTER_LENGTH = 4

    def soft_reset(self):
        """Soft reset the DSP.

        Set and release the corresponding register for resetting.
        """
        self.write(Adau14xx.RESET_REGISTER, int16_to_bytes(0))
        self.write(Adau14xx.RESET_REGISTER, int16_to_bytes(1))
        logger.info("Soft-resetting the DSP.")

    def get_parameter_value(self, address: int, data_format: str) -> Union[float, int, None]:
        """Get a parameter value from a chosen register address.

        Args:
            address (int): The address to look at.
            data_format (str): The data type to return the register in. Can be 'float' or 'int'.

        Returns:
            Union[float, int, None]: Representation of the register content in the specified format.
        """
        data_register = self.read(address, Adau14xx.FIXPOINT_REGISTER_LENGTH)
        data_integer = bytes_to_int32(data_register)

        if "int" == data_format:
            return data_integer

        elif "float" == data_format:
            return frac_8_24_to_float(data_integer)

        else:
            return None

    def set_parameter_value(self, value: Union[float, int], address: int) -> None:
        """Set a parameter value for a chosen register address.

        Args:
            value (float): The value to store in the register
            address (int): The target address
        """
        data_register: Union[bytes, None] = None

        if isinstance(value, float):
            data_register = int32_to_bytes(float_to_frac_8_24(value))

        elif isinstance(value, int):
            data_register = int32_to_bytes(value)

        if data_register is not None:
            self.safeload(address, data_register)

    def set_volume(self, value_db: float, address: int) -> float:
        """Set the volume register at the given address to a certain value in dB.

        Args:
            value_db (float): The volume setting in dB
            address (int): The volume adjustment register address

        Returns:
            float: The new volume in dB.
        """
        # Read current volume and apply adjustment
        value_linear = db_to_linear(value_db)

        # Clamp set volume to safe levels
        clamp(value_linear, 0, 1)

        self.set_parameter_value(value_linear, address)

        logger.info("Set volume to %.2f dB.", linear_to_db(value_linear))

        return linear_to_db(value_linear)

    def adjust_volume(self, adjustment_db: float, address: int) -> float:
        """Adjust the volume register at the given address by a certain value in dB.

        Args:
            adjustment_db (float): The volume adjustment in dB
            address (int): The volume adjustment register address

        Returns:
            float: The new volume in dB.
        """
        # Read current volume and apply adjustment
        current_volume = self.get_parameter_value(address, data_format="float")

        if not isinstance(current_volume, float):
            raise TypeError

        linear_adjustment = db_to_linear(adjustment_db)
        new_volume = current_volume * linear_adjustment

        # Clamp new volume to safe levels
        clamp(new_volume, 0, 1)

        self.set_parameter_value(new_volume, address)

        logger.info(
            "Adjusted volume from %.2f dB to %.2f dB.",
            linear_to_db(current_volume),
            linear_to_db(new_volume),
        )

        return linear_to_db(new_volume)

    def safeload(self, address: int, data: bytes, count: int = 1):
        """Write data to the chip using software safeload.

        Args:
            address (int): Address to write to
            data (bytes): Data to write; multiple parameters should be concatenated
            count (int): Number of 4 byte words to write
        """
        if count > len(Adau14xx.SAFELOAD_DATA_REGISTERS):
            raise IndexError(
                f"Cannot write {count * 4} bytes by means of software safeload, the maximum is "
                f"{len(Adau14xx.SAFELOAD_DATA_REGISTERS) * 4} bytes."
            )
        for register_index, register_address in zip(range(count), Adau14xx.SAFELOAD_DATA_REGISTERS):
            data_buf = data[
                register_index * self.FIXPOINT_REGISTER_LENGTH : (register_index + 1) * self.FIXPOINT_REGISTER_LENGTH
            ]
            self.write(register_address, data_buf)

        # TODO: test if the address is supposed to be shifted down by 1 as old forum posts suggest
        self.write(self.SAFELOAD_ADDRESS_REGISTER, int32_to_bytes(address))
        self.write(self.SAFELOAD_COUNT_REGISTER, int32_to_bytes(count))



"""General definitions for interfacing DSPs."""
import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Type, Union

import gpiozero

from sigmadsp.hardware.base_protocol import BaseProtocol
from sigmadsp.hardware.i2c import I2C
from sigmadsp.hardware.spi import SPI

# A logger for this module
logger = logging.getLogger(__name__)


class SafetyCheckException(Exception):
    """Custom exception for failed DSP safety checks."""


class ConfigurationError(Exception):
    """Custom exception for invalid DSP configuration."""


@dataclass
class Pin:
    """A class that describes a general DSP pin."""

    name: str
    number: int

    def __post_init__(self):
        """Initialize the generic device, based on the configured parameters."""
        self.control = gpiozero.GPIODevice(self.number)


@dataclass
class InputPin(Pin):
    """A class that describes a DSP input pin."""

    pull_up: bool
    active_state: bool
    bounce_time: Union[float, None]

    def __post_init__(self):
        """Initialize the input device, based on the configured parameters."""
        self.control = gpiozero.DigitalInputDevice(self.number, self.pull_up, self.active_state, self.bounce_time)


@dataclass
class OutputPin(Pin):
    """A class that describes a DSP output pin."""

    initial_value: bool
    active_high: bool

    def __post_init__(self):
        """Initialize the output device, based on the configured parameters."""
        self.control = gpiozero.DigitalOutputDevice(self.number, self.active_high, self.initial_value)


class Dsp(ABC):
    """A generic DSP class, to be extended by child classes."""

    protocol_handler: BaseProtocol
    type: str
    protocol: str
    bus: int
    address: int

    def __init__(self, config: dict):
        """Initialize the DSP and set up the protocol handler that talks to it.

        Args:
            config (dict): Configuration settings, from the general configuration file.
        """
        self.config = config
        self.pins: List[Pin] = []
        self.parse_config()

        protocol_handlers: Dict[str, Type[BaseProtocol]] = {
            "i2c": I2C,
            "spi": SPI,
        }

        try:
            handler_class: Type[BaseProtocol] = protocol_handlers[self.protocol]
            self.protocol_handler = handler_class(bus=self.bus, device=self.address)
        except KeyError as e:
            logger.error("Unknown protocol: %s", self.protocol)
            raise ConfigurationError from e

        self.hard_reset()

    def parse_config(self):
        """Parse the configuration file and extract relevant information."""
        try:
            for pin_definition_key in self.config["dsp"]["pins"]:
                pin_definition = self.config["dsp"]["pins"][pin_definition_key]

                if pin_definition["mode"] == "output":
                    output_pin = OutputPin(
                        pin_definition_key,
                        pin_definition["number"],
                        pin_definition["initial_state"],
                        pin_definition["active_high"],
                    )

                    self.add_pin(output_pin)

                elif pin_definition["mode"] == "input":
                    input_pin = InputPin(
                        pin_definition_key,
                        pin_definition["number"],
                        pin_definition["pull_up"],
                        pin_definition["active_state"],
                        pin_definition["bounce_time"],
                    )

                    self.add_pin(input_pin)

        except (KeyError, TypeError):
            logger.info("No DSP pin definitions were found in the configuration file.")

        try:
            self.type = self.config["dsp"]["type"]
            self.protocol = self.config["dsp"]["protocol"]
            self.bus = self.config["dsp"]["bus_number"]
            self.address = self.config["dsp"]["device_address"]
        except KeyError as e:
            logger.error("Key %s missing from the DSP configuration.", e.args[0])
            raise ConfigurationError from e

    def get_pin_by_name(self, name: str) -> Union[Pin, None]:
        """Get a pin by its name.

        Args:
            name (str): The name of the pin.

        Returns:
            Union[Pin, None]: The pin, if one matches, None otherwise.
        """
        for pin in self.pins:
            if pin.name == name:
                return pin

        return None

    def has_pin(self, pin: Pin) -> bool:
        """Check, if a pin is known in the list of pins.

        Args:
            pin (Pin): The pin to look for.

        Returns:
            bool: True, if the pin is known, False otherwise.
        """
        return bool(self.get_pin_by_name(pin.name) is not None)

    def add_pin(self, pin: Pin):
        """Add a pin to the list of pins, if it doesn't exist yet.

        Args:
            pin (Pin): The pin to add.
        """
        if not self.has_pin(pin):
            logger.info("Found DSP pin definition '%s' (%d)", pin.name, pin.number)
            self.pins.append(pin)

    def remove_pin_by_name(self, name: str):
        """Remove a pin, based on its name.

        Args:
            name (str): The pin name.
        """
        pin = self.get_pin_by_name(name)

        if pin:
            self.pins.remove(pin)

    def hard_reset(self, delay: float = 0):
        """Hard reset the DSP.

        Set and release the corresponding pin for resetting.
        """
        pin = self.get_pin_by_name("reset")

        if not pin:
            logger.info("Falling back to soft-resetting the DSP, no hard-reset pin is defined.")
            self.soft_reset()
            return

        logger.info("Hard-resetting the DSP.")

        pin.control.on()
        time.sleep(delay)
        pin.control.off()

    def write(self, address: int, data: bytes):
        """Write data to the DSP using the configured communication handler.

        Args:
            address (int): Address to write to
            data (bytes): Data to write
        """
        self.protocol_handler.write(address, data)

    def read(self, address: int, length: int) -> bytes:
        """Write data to the DSP using the configured communication handler.

        Args:
            address (int): Address to write to
            length (int): Number of bytes to read
        """
        return self.protocol_handler.read(address, length)

    @abstractmethod
    def soft_reset(self):
        """Soft reset the DSP."""

    @abstractmethod
    def safeload(self, address: int, data: bytes, count: int):
        """Write data to the chip using chip-specific safeload.

        Args:
            address (int): Address to write to
            data (bytes): Data to write
            count (int): Number of 4 byte words to write (max. 5)
        """
